@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -47,7 +47,7 @@
 
 #include "../src/dbus-common.h"
 
-#include "logging.h"
+#include "log.h"
 #include "glib-helper.h"
 #include "btio.h"
 
@@ -78,6 +78,7 @@ struct serial_port {
 	char		*uuid;		/* service identification */
 	char		*dev;		/* RFCOMM device name */
 	int		fd;		/* Opened file descriptor */
+	GIOChannel	*io;		/* BtIO channel */
 	guint		listener_id;
 	struct serial_device *device;
 };
@@ -101,26 +102,34 @@ static struct serial_device *find_device(GSList *devices, const char *path)
 static struct serial_port *find_port(GSList *ports, const char *pattern)
 {
 	GSList *l;
+	int channel;
+	char *endptr = NULL;
+
+	channel = strtol(pattern, &endptr, 10);
 
 	for (l = ports; l != NULL; l = l->next) {
 		struct serial_port *port = l->data;
 		char *uuid_str;
 		int ret;
 
-		if (!strcasecmp(port->uuid, pattern))
+		if (port->uuid && !strcasecmp(port->uuid, pattern))
+			return port;
+
+		if (endptr && *endptr == '\0' && port->channel == channel)
 			return port;
 
 		if (port->dev && !strcmp(port->dev, pattern))
 			return port;
 
+		if (!port->uuid)
+			continue;
+
 		uuid_str = bt_name2string(pattern);
 		if (!uuid_str)
 			continue;
 
 		ret = strcasecmp(port->uuid, uuid_str);
-
 		g_free(uuid_str);
-
 		if (ret == 0)
 			return port;
 	}
@@ -134,7 +143,19 @@ static int port_release(struct serial_port *port)
 	int rfcomm_ctl;
 	int err = 0;
 
-	debug("Serial port %s released", port->dev);
+	if (port->id < 0) {
+		if (port->io) {
+			g_io_channel_shutdown(port->io, TRUE, NULL);
+			g_io_channel_unref(port->io);
+			port->io = NULL;
+		} else
+			bt_cancel_discovery(&port->device->src,
+						&port->device->dst);
+
+		return 0;
+	}
+
+	DBG("Serial port %s released", port->dev);
 
 	rfcomm_ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM);
 	if (rfcomm_ctl < 0)
@@ -175,8 +196,7 @@ static void serial_port_free(struct serial_port *port)
 	if (device && port->listener_id > 0)
 		g_dbus_remove_watch(device->conn, port->listener_id);
 
-	if (port->id >= 0)
-		port_release(port);
+	port_release(port);
 
 	g_free(port->uuid);
 	g_free(port);
@@ -194,8 +214,7 @@ static void port_owner_exited(DBusConnection *conn, void *user_data)
 {
 	struct serial_port *port = user_data;
 
-	if (port->id >= 0)
-		port_release(port);
+	port_release(port);
 
 	port->listener_id = 0;
 }
@@ -204,7 +223,7 @@ static void path_unregister(void *data)
 {
 	struct serial_device *device = data;
 
-	debug("Unregistered interface %s on path %s", SERIAL_PORT_INTERFACE,
+	DBG("Unregistered interface %s on path %s", SERIAL_PORT_INTERFACE,
 		device->path);
 
 	devices = g_slist_remove(devices, device);
@@ -221,20 +240,20 @@ static inline DBusMessage *does_not_exist(DBusMessage *msg,
 					const char *description)
 {
 	return g_dbus_create_error(msg, ERROR_INTERFACE ".DoesNotExist",
-				description);
+							"%s", description);
 }
 
 static inline DBusMessage *invalid_arguments(DBusMessage *msg,
 					const char *description)
 {
 	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
-				description);
+							"%s", description);
 }
 
 static inline DBusMessage *failed(DBusMessage *msg, const char *description)
 {
 	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
-				description);
+							"%s", description);
 }
 
 static void open_notify(int fd, int err, struct serial_port *port)
@@ -324,6 +343,9 @@ static void rfcomm_connect_cb(GIOChannel *chan, GError *conn_err,
 	bacpy(&req.dst, &device->dst);
 	req.channel = port->channel;
 
+	g_io_channel_unref(port->io);
+	port->io = NULL;
+
 	sk = g_io_channel_unix_get_fd(chan);
 	port->id = ioctl(sk, RFCOMMCREATEDEV, &req);
 	if (port->id < 0) {
@@ -336,7 +358,7 @@ static void rfcomm_connect_cb(GIOChannel *chan, GError *conn_err,
 
 	port->dev = g_strdup_printf("/dev/rfcomm%d", port->id);
 
-	debug("Serial port %s created", port->dev);
+	DBG("Serial port %s created", port->dev);
 
 	g_io_channel_shutdown(chan, TRUE, NULL);
 
@@ -355,22 +377,138 @@ fail:
 	port->listener_id = 0;
 }
 
+static void get_record_cb(sdp_list_t *recs, int err, gpointer user_data)
+{
+	struct serial_port *port = user_data;
+	struct serial_device *device = port->device;
+	sdp_record_t *record = NULL;
+	sdp_list_t *protos;
+	DBusMessage *reply;
+	GError *gerr = NULL;
+
+	if (!port->listener_id) {
+		reply = NULL;
+		goto failed;
+	}
+
+	if (err < 0) {
+		error("Unable to get service record: %s (%d)", strerror(-err),
+			-err);
+		reply = failed(port->msg, strerror(-err));
+		goto failed;
+	}
+
+	if (!recs || !recs->data) {
+		error("No record found");
+		reply = failed(port->msg, "No record found");
+		goto failed;
+	}
+
+	record = recs->data;
+
+	if (sdp_get_access_protos(record, &protos) < 0) {
+		error("Unable to get access protos from port record");
+		reply = failed(port->msg, "Invalid channel");
+		goto failed;
+	}
+
+	port->channel = sdp_get_proto_port(protos, RFCOMM_UUID);
+
+	sdp_list_foreach(protos, (sdp_list_func_t) sdp_list_free, NULL);
+	sdp_list_free(protos, NULL);
+
+	port->io = bt_io_connect(BT_IO_RFCOMM, rfcomm_connect_cb, port,
+				NULL, &gerr,
+				BT_IO_OPT_SOURCE_BDADDR, &device->src,
+				BT_IO_OPT_DEST_BDADDR, &device->dst,
+				BT_IO_OPT_CHANNEL, port->channel,
+				BT_IO_OPT_INVALID);
+	if (!port->io) {
+		error("%s", gerr->message);
+		reply = failed(port->msg, gerr->message);
+		g_error_free(gerr);
+		goto failed;
+	}
+
+	return;
+
+failed:
+	g_dbus_remove_watch(device->conn, port->listener_id);
+	port->listener_id = 0;
+	g_dbus_send_message(device->conn, reply);
+}
+
+static int connect_port(struct serial_port *port)
+{
+	struct serial_device *device = port->device;
+	uuid_t uuid;
+	int err;
+
+	if (!port->uuid)
+		goto connect;
+
+	err = bt_string2uuid(&uuid, port->uuid);
+	if (err < 0)
+		return err;
+
+	sdp_uuid128_to_uuid(&uuid);
+
+	return bt_search_service(&device->src, &device->dst, &uuid,
+				get_record_cb, port, NULL);
+
+connect:
+	port->io = bt_io_connect(BT_IO_RFCOMM, rfcomm_connect_cb, port,
+				NULL, NULL,
+				BT_IO_OPT_SOURCE_BDADDR, &device->src,
+				BT_IO_OPT_DEST_BDADDR, &device->dst,
+				BT_IO_OPT_CHANNEL, port->channel,
+				BT_IO_OPT_INVALID);
+	if (port->io)
+		return 0;
+
+	return -errno;
+}
+
+static struct serial_port *create_port(struct serial_device *device,
+					const char *uuid, uint8_t channel)
+{
+	struct serial_port *port;
+
+	port = g_new0(struct serial_port, 1);
+	port->uuid = g_strdup(uuid);
+	port->channel = channel;
+	port->device = device;
+	port->id = -1;
+	port->fd = -1;
+
+	device->ports = g_slist_append(device->ports, port);
+
+	return port;
+}
+
 static DBusMessage *port_connect(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
 	struct serial_device *device = user_data;
 	struct serial_port *port;
-	const char *uuid;
-	GIOChannel *io;
-	GError *err = NULL;
+	const char *pattern;
+	int err;
 
-	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &uuid,
+	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &pattern,
 						DBUS_TYPE_INVALID) == FALSE)
 		return NULL;
 
-	port = find_port(device->ports, uuid);
-	if (!port)
-		return does_not_exist(msg, "Does not match");
+	port = find_port(device->ports, pattern);
+	if (!port) {
+		char *endptr = NULL;
+		int channel;
+
+		channel = strtol(pattern, &endptr, 10);
+		if ((endptr && *endptr != '\0') || channel < 1 || channel > 30)
+			return does_not_exist(msg, "Does not match");
+
+		port = create_port(device, NULL, channel);
+	}
 
 	if (port->listener_id)
 		return failed(msg, "Port already in use");
@@ -381,25 +519,17 @@ static DBusMessage *port_connect(DBusConnection *conn,
 						NULL);
 	port->msg = dbus_message_ref(msg);
 
-	io = bt_io_connect(BT_IO_RFCOMM, rfcomm_connect_cb, port,
-				NULL, &err,
-				BT_IO_OPT_SOURCE_BDADDR, &device->src,
-				BT_IO_OPT_DEST_BDADDR, &device->dst,
-				BT_IO_OPT_CHANNEL, port->channel,
-				BT_IO_OPT_INVALID);
-	if (!io) {
+	err = connect_port(port);
+	if (err < 0) {
 		DBusMessage *reply;
 
-		error("%s", err->message);
+		error("%s", strerror(-err));
 		g_dbus_remove_watch(conn, port->listener_id);
 		port->listener_id = 0;
-		reply = failed(msg, err->message);
-		g_error_free(err);
+		reply = failed(msg, strerror(-err));
 		return reply;
 	}
 
-	g_io_channel_unref(io);
-
 	return NULL;
 }
 
@@ -426,8 +556,7 @@ static DBusMessage *port_disconnect(DBusConnection *conn,
 	if (!g_str_equal(owner, caller))
 		return failed(msg, "Operation not permited");
 
-	if (port->id >= 0)
-		port_release(port);
+	port_release(port);
 
 	g_dbus_remove_watch(conn, port->listener_id);
 	port->listener_id = 0;
@@ -463,7 +592,7 @@ static struct serial_device *create_serial_device(DBusConnection *conn,
 		return NULL;
 	}
 
-	debug("Registered interface %s on path %s",
+	DBG("Registered interface %s on path %s",
 		SERIAL_PORT_INTERFACE, path);
 
 	return device;
